/*jshint unused:true */
/*
 * This file contains a widget to change workspace names.
 * It can be run by gjs-console without worrying about non-standard
 * imports (in particular, Meta). This makes it compatible with 3.2+.
 *
 * Since gjs-console can't get imports.gi.Meta by default we can't just use
 * Meta.prefs_(get|set)_workspace_names
 * to do all the work for us, so instead we look at the gconf/dconf keys
 * for workspace names (/apps/metacity/workspace_names and
 * org.gnome.desktop.wm.preferences.workspace-names) and modify those.
 *
 * It works in 3.4+ in the prefs widget and also from 3.2 if you run it like so:
 * gjs -I ./ -c "const Main = imports.prefs; Main.main();"
 */
const GConf = imports.gi.GConf;
const Gio = imports.gi.Gio;
const GObject = imports.gi.GObject;
const Gtk = imports.gi.Gtk;
const Lang = imports.lang;
const Signals = imports.signals;

// in GNOME 3.2 GObject.TYPE_STRING doesn't exist.
const G_TYPE_STRING = (GObject.TYPE_STRING ||
                       GObject.type_from_name('gchararray'));

const MAX_REASONABLE_WORKSPACES = 36; // hard-coded into mutter/metacity

const GCONF_PATH = '/apps/metacity/workspace_names';
const DCONF_PATH = 'org.gnome.desktop.wm.preferences';
const DCONF_KEY_WSNAMES = 'workspace-names';

/*********************
 *     SETTINGS      *
 *********************/
/* A class MetaSettings that works in GNOME 3.2 (gconf) or GNOME 3.4 (dconf).
 * It can be used to query/set workspace names and listen to workspace
 * name changes.
 * We only need it because imports.gi.Meta is not available from without
 * the gnome-shell compositor process.
 *
 * It provides two signals:
 * * workspace-names-changed : fired when workspace name(s) change.
 * * workspace-name-changed: fired once per workspace that has its name
 *   changed, giving the index of the new names.
 *
 * Sometimes workspace-names-changed can fire once to cover multiple
 * simultaneous workspace name changes (only in GNOME 3.4; in GNOME 3.2
 * the two are synonomous).
 */
function MetaSettings() {
    this._init.apply(this, arguments);
}

try {
    // this will fail in GNOME 3.2 because you have to use 'schema' not
    // 'schema_id'.
    let settings = new Gio.Settings({
        schema_id: DCONF_PATH
    });
    settings = null;

    // GNOME 3.4
    MetaSettings.prototype = {
        _init: function () {
            this._settings = new Gio.Settings({
                schema_id: DCONF_PATH
            });
            this._names = [];
            this._nameChangedIDs = [];

            this._loadWorkspaceNames();
            this._changedNamesID = this._settings.connect(
                'changed::' + DCONF_KEY_WSNAMES,
                Lang.bind(this, this._onWorkspaceNamesChanged)
            );
        },

        _loadWorkspaceNames: function () {
            this._names = this._settings.get_strv(DCONF_KEY_WSNAMES);
        },

        _onWorkspaceNamesChanged: function () {
            this.emit('workspace-names-changed');
            let oldnames = this._names;
            this._loadWorkspaceNames();
            // this might be a bit naive, but we simply fire a
            // workspace-name-changed signal for *every* workspace that doesn't
            // agree with what it used to be. This can cause a whole bunch
            // of signals being fired at once.
            for (let i = 0; i < Math.min(MAX_REASONABLE_WORKSPACES,
                        Math.max(oldnames.length, this._names.length)); ++i) {
                if (this._names[i] !== oldnames[i]) {
                    let newname = this._translateName(i);
                    this.emit('workspace-name-changed', i, newname);
                    this.emit('workspace-' + i + '-name-changed', newname);
                }
            }
        },

        // Will fire 'workspace-name-changed'. i is 0-based.
        change_workspace_name: function (i, newname) {
            // TODO: translations of 'Workspace ' ?
            if (newname === 'Workspace ' + (i + 1).toString()) {
                newname = '';
            }
            if (i < 0 || i >= MAX_REASONABLE_WORKSPACES ||
                this._names[i] == newname) {
                return;
            }

            this._names[i] = newname;
            // note if i > this._names.length we'll get weird uninitialised
            // elements in between and .map and .set_strv won't work.
            for (let j = 0; j < this._names.length; ++j) {
                if (!this._names[j]) {
                    this._names[j] = '';
                }
            }
            this._settings.set_strv(DCONF_KEY_WSNAMES, this._names);
        },

        destroy: function () {
            this.disconnectAll();
            this._names = null;
            this._settings.disconnect(this._changedNamesID);
            this._settings = null;
        }
    };
} catch (err) {
    // GNOME 3.2
    MetaSettings.prototype = {
        _init: function () {
            this._client = GConf.Client.get_default();
            this._names = [];
            this._nameChangedIDs = [];

            this._loadWorkspaceNames();
            // you must add the directory for it to be watched
            this._client.add_dir(GCONF_PATH,
                    GConf.ClientPreloadType.PRELOAD_ONELEVEL);
        },

        _loadWorkspaceNames: function () {
            let entries = this._client.all_entries(GCONF_PATH);
            // NOTE: on GNOME 3.2, name_%i is 1-based.
            for (let i = 0; i < entries.length; ++i) {
                let entry = entries[i],
                    key = entry.key,
                    wsI = key.match(/name_([0-9]+)$/)[1];
                this._names[wsI - 1] = entry.value.get_string();

                // connect to name change: emit 'workspace-name-changed'
                // with the workspace *index* and new name.
                this._nameChangedIDs.push(this._client.notify_add(key,
                    Lang.bind(this, function (client, id, entry) {
                        let newname = entry.value.get_string();
                        this._names[wsI - 1] = newname;
                        this.emit('workspace-name-changed', wsI, newname);
                        this.emit('workspace-' + (wsI - 1) + '-name-changed',
                            newname);
                        // the way gconf is set up you don't get simultaneous
                        // workspace name changes so we can emit them all on
                        // a per-workspace basis.
                        this.emit('workspace-names-changed');
                    })
                ));
            }
        },

        // Will fire 'workspace-name-changed'. i is 0-based.
        change_workspace_name: function (i, newname) {
            if (newname === 'Workspace ' + (i + 1).toString()) {
                newname = '';
            }
            if (i < 0 || i >= MAX_REASONABLE_WORKSPACES ||
                this._names[i] == newname) {
                return;
            }
            this._client.set_string(GCONF_PATH + '/name_' + (i + 1).toString(),
                newname);
        },

        destroy: function () {
            this.disconnectAll();
            let i = this._nameChangedIDs.length;
            while (i--) {
                this._client.notify_remove(this._nameChangedIDs.pop());
            }
            this._client.remove_dir(GCONF_PATH);
            this._names = null;
            this._nameChangedIDs = null;
            this._client = null;
        }
    };
}

MetaSettings.prototype._translateName = function (i) {
    return this._names[i] || ('Workspace ' + (i + 1).toString());
};

// 0-based workspace index.
MetaSettings.prototype.get_workspace_name = function (i) {
    if (i < 0 || i >= MAX_REASONABLE_WORKSPACES) {
        return null;
    }
    return this._translateName(i);
};
Signals.addSignalMethods(MetaSettings.prototype);


/**************************
 * Workspace names widget *
 **************************/
function WorkspaceNamesWidget() {
    this._init.apply(this, arguments);
}

WorkspaceNamesWidget.prototype = {
    _init: function () {
        // settings
        this._settings = new MetaSettings();

        let area;
        this.widget = area = new Gtk.Box({
            orientation: Gtk.Orientation.VERTICAL,
            spacing: 10
        });

        // create a label with the title
        this.title = new Gtk.Label({
            label: "Change workspace names",
            hexpand: true,
            halign: Gtk.Align.START
        });
        this.title.set_line_wrap(true);
        area.add(this.title);

        // create the tree view holding the workspaces
        this.workspacesTreeView = new Gtk.TreeView({
            hexpand: true,
            vexpand: true,
            headers_visible: false
        });

        // add it to the dialog in a scroll window
        let sw = new Gtk.ScrolledWindow({
            hscrollbar_policy: Gtk.PolicyType.AUTOMATIC,
            vscrollbar_policy: Gtk.PolicyType.AUTOMATIC,
            hexpand: true,
            vexpand: true
        });

        sw.add_with_viewport(this.workspacesTreeView);
        sw.set_min_content_height(300);
        sw.show_all();
        area.add(sw);

        this._store = new Gtk.ListStore();
        this._store.set_column_types([G_TYPE_STRING]); // workspace name
        this.workspacesTreeView.set_model(this._store);

        let cell = new Gtk.CellRendererText({editable: true}),
            column = new Gtk.TreeViewColumn({
                title: "workspace"
            });
        column.pack_start(cell, true);
        column.add_attribute(cell, 'text', 0); // 0th column is the text
        this.workspacesTreeView.append_column(column);
        cell.connect('edited', Lang.bind(this, this._onWorkspaceNameEdited));

        /* Listen for workspace name change: this will have to be
         * via the settings
         */
        this._settings.connect('workspace-names-changed', Lang.bind(this,
            this._updateWorkspacesModel));
        this._updateWorkspacesModel();
    },

    /* pass through some functions */
    show_all: function () {
        this.widget.show_all();
    },

    destroy: function () {
        this._settings.disconnectAll();
        this._settings.destroy();
        this._settings = null;

        this.workspacesTreeView.destroy();
    },

    //// callbacks ////
    _onWorkspaceNameEdited: function (cell, path, new_text) {
        let idx = parseInt(path, 10);
        this._settings.change_workspace_name(idx, new_text);
    },

    _updateWorkspacesModel: function () {
        this._store.clear();
        for (let i = 0; i < MAX_REASONABLE_WORKSPACES; ++i) {
            let iter = this._store.append();
            this._store.set_value(iter, 0,
                this._settings.get_workspace_name(i));
        }
    }
};

function WorkspaceNamesDialog() {
    this._init.apply(this, arguments);
}

WorkspaceNamesDialog.prototype = {
    _init: function () {
        // build the UI
        this._widget = new WorkspaceNamesWidget();

        let dialog = this.dialog = new Gtk.Dialog({
            title: this._widget.title.label
        });

        // centre buttons in the dialog
        dialog.get_action_area().set_layout(Gtk.ButtonBoxStyle.CENTER);

        // add some padding between the dialog and its contents
        dialog.set_border_width(10);
        // add the contents
        dialog.get_content_area().add(this._widget.widget);

        dialog.add_button(Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE);
        dialog.set_default_response(Gtk.ResponseType.CLOSE);

        // pass through some signals
        dialog.connect('response', Lang.bind(this, this.destroy));
        dialog.connect('close', Lang.bind(this, this.destroy));

        /* Finishing touches */
        // if they have gnome-panel installed this will do the trick
        this.dialog.set_icon_name('gnome-panel-workspace-switcher');

    },

    /* pass through some functions */
    show_all: function () {
        this.dialog.show_all();
    },

    destroy: function () {
        this._widget.destroy();
        this.dialog.destroy();

        this.dialog = null;
        Gtk.main_quit(); // quit the mainloop
    }
};

function main() {
    Gtk.init(null);
    let dialog = new WorkspaceNamesDialog();
    dialog.show_all();

    Gtk.main();
}

function init() {
}

function buildPrefsWidget() {
    let widget = new WorkspaceNamesWidget();
    widget.show_all();

    return widget.widget;
}
